"""
KarambaMetar (c) 2003 Jan Finell
Superkaramba weather theme using the pymetar module to fetch the weather data.

Thanks to Tobias Klausman for the pymetar module.
See the pymetar.py module for more information on it.

Weather icons set from weather.com.
To change the icon set just overwrite the icons in weather_icons/ directory.
(note! You can use the icons provided with liquid-weather ;)).
Scale down the large icons to 45x45 otherwise it will look ugly! The small
icons should be 22x22.

Get you local 4-character ICAO station code from:
http://www.nws.noaa.gov/tg/siteloc.shtml

### LICENSE ########################################################
This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU General Public License for more details.
###################################################################

Changes:
  0.6 : Showing different icons during the night.
        Fixed bug in the pymetar module.
"""

__author__  = 'Jan Finell'
__version__ = '0.6'

#
# standard modules
#
import time

#
# extension modules
#
import karamba, pymetar

#
# variables & constants
#
class Var:
    """
    Helper class for keeping runtime variables
    """
    def __init__(self, **kws):
        self.__dict__ = kws

#stationCode EFTU is Turku, Finland

VAR = Var(weather=None, shown=False, utime=None, timezone=2,
          stationCode='EFTU', timezonePID=None, stationCodePID=None)

METAR_TIME_FORMAT = '%Y.%m.%d %H%M %Z'
LOCAL_TIME_FORMAT = '%d %b %H:%M'
IMAGE_X, IMAGE_Y  = (670,16)
DETAILS_X, DETAILS_Y = (670,16)

NIGHT_PIXMAPNO = {'sun' : 31, 'suncloud' : 27, 'cloud': 26, 'rain': 45,
                  'snow': 46 , 'fog' : 29, 'storm': 47}           
#
# metar functions
#
def fetch_weather_report(station_code):
    try:
        print '-- fetching report for %s --' % station_code
        rf = pymetar.ReportFetcher(station_code)
        rep = rf.FetchReport()
        print '-- parsing report --'
        rp = pymetar.ReportParser()
        wr = rp.ParseReport(rep)
    except Exception, e:
        print 'Error fetching weather report (%s)' % e
        wr = None
    return wr

def get_localtime(time_string):
    try:
        tt = time.strptime(time_string, METAR_TIME_FORMAT)
        tsec = time.mktime(tt)
        tsec += int(VAR.timezone)*3600
        lt = time.strftime(LOCAL_TIME_FORMAT, time.localtime(tsec))
    except Exception, e:
        print 'Error parsing time (%s)' % e
        lt = 'n.a.'
    return lt

def is_night_time(localtime):
    try:
        tt = time.strptime(localtime, LOCAL_TIME_FORMAT)
        # between 7pm and 5am is night-time ;)
        return 19 <= tt[3] or tt[3] <= 5
    except:
        return False
        
#
# Karamba functions
#
def initWidget(widget):
    VAR.stationCode = karamba.readConfigEntry(widget,'stationCode') or \
                      VAR.stationCode
    VAR.timezone = karamba.readConfigEntry(widget,'timezone') or \
                   VAR.timezone
    VAR.image = karamba.createImage(widget, IMAGE_X, IMAGE_Y,
                                    'weather_icons/small/25.png')
    karamba.attachClickArea(widget, karamba.getThemeText(widget,'MASK'))
    karamba.addMenuConfigOption(widget, 'stationcode', 'Change station code')
    karamba.setMenuConfigOption(widget, 'stationcode', '0')
    karamba.addMenuConfigOption(widget, 'timezone','Change timezone')
    karamba.setMenuConfigOption(widget, 'timezone', '0')
    karamba.redrawWidget(widget)
    widgetUpdated(widget)

def widgetUpdated(widget):
    if VAR.utime is None or time.time()-VAR.utime > 3000:
        VAR.weather = fetch_weather_report(VAR.stationCode)
        VAR.utime = time.time()
    _update(widget)

def meterClicked(widget, meter, button):
    if VAR.weather:
        if VAR.shown:
            _hide_details(widget)
        else:
            _show_details(widget)
        karamba.redrawWidget(widget)
        VAR.shown = not VAR.shown

def menuOptionChanged(widget, key, value):
    if key == 'stationcode':
        cmd = ['kdialog', '--title', 'Change station code', '--inputbox',
               'Please enter your 4-character station ICAO ID '\
               '(from http://www.nws.noaa.gov/tg/siteloc.shtml):',
               VAR.stationCode]
        karamba.setMenuConfigOption(widget, 'stationcode', '0')
        VAR.stationCodePID = karamba.executeInteractive(widget, cmd)
    elif key == 'timezone':
        cmd = ['kdialog', '--title', 'Change timezone', '--inputbox',
               'Please enter your timezone (eg. 2 for GMT+2):',
               str(VAR.timezone)]
        karamba.setMenuConfigOption(widget, 'timezone', '0')
        VAR.timezonePID = karamba.executeInteractive(widget, cmd)
               
def commandOutput(widget, pid, output):
    if pid == VAR.timezonePID:
        try:
            timezone = int(output)
            karamba.writeConfigEntry(widget,'timezone', str(timezone))
            VAR.timezone = timezone
        except:
            cmd = ['kdialog', '--title', 'Invalid Input',
                   '--error', 'Given timezone is not valid!']
            karamba.executeInteractive(widget, cmd)
            return
    elif pid == VAR.stationCodePID:
        stationCode = output.replace('\n','').strip()
        if len(stationCode) == 4:
            karamba.writeConfigEntry(widget,'stationCode', stationCode)
            VAR.stationCode = stationCode
        else:
            cmd = ['kdialog', '--title', 'Invalid Input',
                   '--error', 'Given station code is not valid!']
            karamba.executeInteractive(widget, cmd)
            return
    VAR.utime = None # need to reset otherwise it will not upda
    widgetUpdated(widget)

    
#
# protected functions
#
def _update(widget):
    temp = wind = ''
    pixmapno = 25
    if VAR.weather:
        wr = VAR.weather
        temp = '%s %sC' % (round(wr.getTemperatureCelsius(),1), chr(176))
        wind = '%d m/s %s' % (wr.getWindSpeed(), wr.getWindCompass())
        pixmapno = wr.getPixmapNo() or 25
        if is_night_time(get_localtime(wr.getTime())):        
            pixmapno = NIGHT_PIXMAPNO.get(wr.getPixmap(), pixmapno)
    karamba.changeText(widget, karamba.getThemeText(widget, 'TEMP'), temp)
    karamba.changeText(widget, karamba.getThemeText(widget, 'WIND'), wind)
    karamba.deleteImage(widget, VAR.image)
    VAR.image = karamba.createImage(widget, IMAGE_X, IMAGE_Y,
                                    'weather_icons/small/%s.png' % pixmapno)
    karamba.redrawWidget(widget)

def _show_details(widget):
    bg = karamba.createImage(widget, DETAILS_X, DETAILS_Y,
                              'images/details_bg.png')
    for i in range(115, 35, 5):
        karamba.resizeImage(widget, bg, 180, i)
        karamba.redrawWidget(widget)
    city = wtime = weather = conditions = ''
    pixmapno = 25
    if VAR.weather:
        wr = VAR.weather
        city = wr.getStationCity()
        wtime = get_localtime(wr.getTime())
        weather = wr.getWeather()
        conditions = wr.getConditions() or '-'
        pixmapno  = wr.getPixmapNo() or 25
        if is_night_time(wtime):        
            pixmapno = NIGHT_PIXMAPNO.get(wr.getPixmap(), pixmapno)
    img = karamba.createImage(widget, DETAILS_X-100, IMAGE_Y-33,
                              'weather_icons/large/%s.png' % pixmapno)
    titleW = karamba.createText(widget, DETAILS_X-220, DETAILS_Y-15,
                                170, 10, '%s %s' % (city, wtime))
    karamba.changeTextSize(widget, titleW, 11)
    karamba.changeTextFont(widget, titleW, 'verdana')
    karamba.changeTextColor(widget, titleW, 214, 214, 214)
    karamba.changeTextShadow(widget, titleW, 1)
    txtW = karamba.createText(widget, DETAILS_X-15, DETAILS_Y-15,
                               170, 15, '%s (%s)' % (weather, conditions))
    karamba.changeTextSize(widget, txtW, 11)
    karamba.changeTextFont(widget, txtW, 'verdana')
    karamba.changeTextColor(widget, txtW, 5, 64, 78)
    VAR.detailsW = (bg, img, titleW, txtW)

def _hide_details(widget):
    karamba.hideImage(widget, VAR.detailsW[0])
    karamba.hideImage(widget, VAR.detailsW[1])
    karamba.hideText(widget, VAR.detailsW[2])
    karamba.hideText(widget, VAR.detailsW[3])
    karamba.deleteImage(widget, VAR.detailsW[0])
    karamba.deleteImage(widget, VAR.detailsW[1])
    karamba.deleteText(widget, VAR.detailsW[2])
    karamba.deleteText(widget, VAR.detailsW[3])
    VAR.detailsW = None 
    
