# Copyright (C) 2002  Tobias Klausmann
# Modified by Jerome Alet
# Code contributed by Jerome Alet and Davide Di Blasi
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.        See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#
# By reading this code you agree not to ridicule the author =)

# When reading this code you'll noctice that not everything is parsed
# from the raw METAR info as supplied with the report. Before you start
# flaming the authors, please read the FHM-1 spec and guess how hairy a
# parser for that format would be. As a side note, the reference 
# implementation of the parser is written in... FORTRAN. Now stop crying.

import fpformat
import math
import re
import string
import urllib2
import types

__author__ = "klausman-pymetar@schwarzvogel.de"

__version__ = "0.6b"

__doc__ = """Pymetar v%s (c) 2002, 2003 Tobias Klausman

Pymetar is a python module and command line tool designed to fetch Metar
reports from the NOAA (http://www.noaa.gov) and allow access to the
included weather information.

This program is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published by the
Free Software Foundation; either version 2 of the License, or (at your
option) any later version.

This program is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License along
with this program; if not, write to the Free Software Foundation, Inc.,
59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.

Modified:

- J.Finell : * importing missing modules (types)
             * added missing NetworkException class
             * added pixmap-no information to the WeatherConditions dictionary
               and to the WeatherReport class
             * added conditions attribute to the WeatherReport class

Please e-mail bugs to: %s""" % (__version__, __author__)


CLOUD_RE_STR= r"^(CLR|SKC|BKN|SCT|FEW|OVC)([0-9]{3})?$"
COND_RE_STR = r"^(-|\\+)?(VC|MI|BC|PR|TS|BL|SH|DR|FZ)?(DZ|RA|SN|SG|IC|PE|GR|GS|UP|BR|FG|FU|VA|SA|HZ|PY|DU|SQ|SS|DS|PO|\\+?FC)$"

class EmptyReportException(Exception):
    def __init__(self, args=None):
        self.args=args

class EmptyIDException(Exception):
    def __init__(self, args=None):
        self.args=args

class NetworkException(Exception):
    def __init__(self, args=None):
        self.args=args

# What a boring list to type !
#
# It seems the NOAA doesn't want to return plain text, but considering the
# format of their response, this is not to save bandwidth :-)
#
_WeatherConditions = {
                      "DZ" : ("Drizzle", "rain", {
                               "" : (  "Moderate drizzle", 9),
                               "-" : ( "Light drizzle", 9),
                               "+" : ( "Heavy drizzle", 9),
                               "VC" : ("Drizzle in the vicinity", 9),
                               "MI" : ("Shallow drizzle", 9),
                               "BC" : ("Patches of drizzle", 9),
                               "PR" : ("Partial drizzle", 9),
                               "TS" : ("Thunderstorm",  9),
                               "BL" : ("Windy drizzle", 9),
                               "SH" : ("Showers", 9),
                               "DR" : ("Drifting drizzle", 9),
                               "FZ" : ("Freezing drizzle", 8),
                             }),
                      "RA" : ("Rain", "rain", {
                               "" : (  "Moderate rain", 12),
                               "-" : ( "Light rain",11),
                               "+" : ( "Heavy rain",40),
                               "VC" : ("Rain in the vicinity",11),
                               "MI" : ("Shallow rain",11),
                               "BC" : ("Patches of rain",11),
                               "PR" : ("Partial rainfall",11),
                               "TS" : ("Thunderstorm", 17),
                               "BL" : ("Blowing rainfall",1),
                               "SH" : ("Rain showers",11),
                               "DR" : ("Drifting rain",11),
                               "FZ" : ("Freezing rain",10),
                             }),
                      "SN" : ("Snow", "snow", {
                               "" : (  "Moderate snow", 14),
                               "-" : ( "Light snow",13),
                               "+" : ( "Heavy snow",42),
                               "VC" : ("Snow in the vicinity",14),
                               "MI" : ("Shallow snow",14),
                               "BC" : ("Patches of snow",16),
                               "PR" : ("Partial snowfall",14),
                               "TS" : ("Snowstorm",43),
                               "BL" : ("Blowing snowfall",43),
                               "SH" : ("Snowfall showers",14),
                               "DR" : ("Drifting snow",14),
                               "FZ" : ("Freezing snow",14),
                             }),
                      "SG" : ("Snow grains", "snow", {
                               "" : (  "Moderate snow grains",15),
                               "-" : ( "Light snow grains",15),
                               "+" : ( "Heavy snow grains",15),
                               "VC" : ("Snow grains in the vicinity",15),
                               "MI" : ("Shallow snow grains",15),
                               "BC" : ("Patches of snow grains",15),
                               "PR" : ("Partial snow grains",15),
                               "TS" : ("Snowstorm", 15),
                               "BL" : ("Blowing snow grains",15),
                               "SH" : ("Snow grain showers",15),
                               "DR" : ("Drifting snow grains",15),
                               "FZ" : ("Freezing snow grains",15),
                             }),
                      "IC" : ("Ice crystals", "snow", {
                               "" : (  "Moderate ice crystals",15),
                               "-" : ( "Few ice crystals",15),
                               "+" : ( "Heavy ice crystals",15),
                               "VC" : ("Ice crystals in the vicinity",15),
                               "BC" : ("Patches of ice crystals",15),
                               "PR" : ("Partial ice crystals",15),
                               "TS" : ("Ice crystal storm", 15),
                               "BL" : ("Blowing ice crystals",15),
                               "SH" : ("Showers of ice crystals",15),
                               "DR" : ("Drifting ice crystals",15),
                               "FZ" : ("Freezing ice crystals",15),
                             }),
                      "PE" : ("Ice pellets", "snow", {
                               "" : (  "Moderate ice pellets",15),
                               "-" : ( "Few ice pellets",15),
                               "+" : ( "Heavy ice pellets",15),
                               "VC" : ("Ice pellets in the vicinity",15),
                               "MI" : ("Shallow ice pellets",15),
                               "BC" : ("Patches of ice pellets",15),
                               "PR" : ("Partial ice pellets",15),
                               "TS" : ("Ice pellets storm", 15),
                               "BL" : ("Blowing ice pellets",15),
                               "SH" : ("Showers of ice pellets",15),
                               "DR" : ("Drifting ice pellets",15),
                               "FZ" : ("Freezing ice pellets",15),
                             }),
                      "GR" : ("Hail", "rain", {
                               "" : (  "Moderate hail", 6),
                               "-" : ( "Light hail",6),
                               "+" : ( "Heavy hail",18),
                               "VC" : ("Hail in the vicinity",6),
                               "MI" : ("Shallow hail",6),
                               "BC" : ("Patches of hail",6),
                               "PR" : ("Partial hail",6),
                               "TS" : ("Hailstorm", 6),
                               "BL" : ("Blowing hail",6),
                               "SH" : ("Hail showers",6),
                               "DR" : ("Drifting hail",6),
                               "FZ" : ("Freezing hail",6),
                             }),
                      "GS" : ("Small hail", "rain", {
                               "" : (  "Moderate small hail",6),
                               "-" : ( "Light small hail",6),
                               "+" : ( "Heavy small hail",6),
                               "VC" : ("Small hail in the vicinity",6),
                               "MI" : ("Shallow small hail",6),
                               "BC" : ("Patches of small hail",6),
                               "PR" : ("Partial small hail",6),
                               "TS" : ("Small hailstorm", 6),
                               "BL" : ("Blowing small hail",6),
                               "SH" : ("Showers of small hail",6),
                               "DR" : ("Drifting small hail",6),
                               "FZ" : ("Freezing small hail",6),
                             }),
                      "UP" : ("Precipitation", "rain", {
                               "" : (  "Moderate precipitation",11),
                               "-" : ( "Light precipitation",11),
                               "+" : ( "Heavy precipitation",11),
                               "VC" : ("Precipitation in the vicinity",11),
                               "MI" : ("Shallow precipitation",11),
                               "BC" : ("Patches of precipitation",11),
                               "PR" : ("Partial precipitation",11),
                               "TS" : ("Unknown thunderstorm", 11),
                               "BL" : ("Blowing precipitation",11),
                               "SH" : ("Showers, type unknown",11),
                               "DR" : ("Drifting precipitation",11),
                               "FZ" : ("Freezing precipitation",11),
                             }),
                      "BR" : ("Mist", "fog", {
                               "" : (  "Moderate mist", 20),
                               "-" : ( "Light mist", 20),
                               "+" : ( "Thick mist", 20),
                               "VC" : ("Mist in the vicinity", 20),
                               "MI" : ("Shallow mist", 20),
                               "BC" : ("Patches of mist", 20),
                               "PR" : ("Partial mist", 20),
                               "BL" : ("Mist with wind", 20),
                               "DR" : ("Drifting mist", 20),
                               "FZ" : ("Freezing mist", 20),
                             }),
                      "FG" : ("Fog", "fog", {
                               "" : (  "Moderate fog", 20),
                               "-" : ( "Light fog", 20),
                               "+" : ( "Thick fog", 20),
                               "VC" : ("Fog in the vicinity", 20),
                               "MI" : ("Shallow fog", 20),
                               "BC" : ("Patches of fog", 20),
                               "PR" : ("Partial fog", 20),
                               "BL" : ("Fog with wind", 20),
                               "DR" : ("Drifting fog", 20),
                               "FZ" : ("Freezing fog", 20)
                             }),
                      "FU" : ("Smoke", "fog", {
                               "" : (  "Moderate smoke",22),
                               "-" : ( "Thin smoke",22),
                               "+" : ( "Thick smoke",22),
                               "VC" : ("Smoke in the vicinity",22),
                               "MI" : ("Shallow smoke",22),
                               "BC" : ("Patches of smoke",22),
                               "PR" : ("Partial smoke",22),
                               "TS" : ("Smoke w/ thunders",22), 
                               "BL" : ("Smoke with wind",22),
                               "DR" : ("Drifting smoke",22),
                             }),
                      "VA" : ("Volcanic ash", "fog", {
                               "" : (  "Moderate volcanic ash",22),
                               "+" : ( "Thick volcanic ash",22),
                               "VC" : ("Volcanic ash in the vicinity",22),
                               "MI" : ("Shallow volcanic ash",22),
                               "BC" : ("Patches of volcanic ash",22),
                               "PR" : ("Partial volcanic ash",22),
                               "TS" : ("Volcanic ash w/ thunders", 22),
                               "BL" : ("Blowing volcanic ash",22),
                               "SH" : ("Showers of volcanic ash",22),
                               "DR" : ("Drifting volcanic ash",22),
                               "FZ" : ("Freezing volcanic ash",22),
                             }),
                      "SA" : ("Sand", "fog", {
                               "" : (  "Moderate sand",19),
                               "-" : ( "Light sand",19),
                               "+" : ( "Heavy sand",19),
                               "VC" : ("Sand in the vicinity",19),
                               "BC" : ("Patches of sand",19),
                               "PR" : ("Partial sand",19),
                               "BL" : ("Blowing sand",19),
                               "DR" : ("Drifting sand",19),
                             }),
                      "HZ" : ("Haze", "fog", {
                               "" : (  "Moderate haze",21),
                               "-" : ( "Light haze",21),
                               "+" : ( "Thick haze",21),
                               "VC" : ("Haze in the vicinity",21),
                               "MI" : ("Shallow haze",21),
                               "BC" : ("Patches of haze",21),
                               "PR" : ("Partial haze",21),
                               "BL" : ("Haze with wind",21),
                               "DR" : ("Drifting haze",21),
                               "FZ" : ("Freezing haze",21),
                             }),
                      "PY" : ("Sprays", "fog", {
                               "" : (  "Moderate sprays",21),
                               "-" : ( "Light sprays",21),
                               "+" : ( "Heavy sprays",21),
                               "VC" : ("Sprays in the vicinity",21),
                               "MI" : ("Shallow sprays",21),
                               "BC" : ("Patches of sprays",21),
                               "PR" : ("Partial sprays",21),
                               "BL" : ("Blowing sprays",21),
                               "DR" : ("Drifting sprays",21),
                               "FZ" : ("Freezing sprays",21),
                             }),
                      "DU" : ("Dust", "fog", {
                               "" : (  "Moderate dust",19),
                               "-" : ( "Light dust",19),
                               "+" : ( "Heavy dust",19),
                               "VC" : ("Dust in the vicinity",19),
                               "BC" : ("Patches of dust",19),
                               "PR" : ("Partial dust",19),
                               "BL" : ("Blowing dust",19),
                               "DR" : ("Drifting dust",19),
                             }),
                      "SQ" : ("Squall", "storm", {
                               "" : (  "Moderate squall",23),
                               "-" : ( "Light squall",23),
                               "+" : ( "Heavy squall",23),
                               "VC" : ("Squall in the vicinity",23),
                               "PR" : ("Partial squall",23),
                               "TS" : ("Thunderous squall",23),
                               "BL" : ("Blowing squall",23),
                               "DR" : ("Drifting squall",23),
                               "FZ" : ("Freezing squall",23),
                             }),
                      "SS" : ("Sandstorm", "fog", {
                               "" : (  "Moderate sandstorm",19),
                               "-" : ( "Light sandstorm",19),
                               "+" : ( "Heavy sandstorm",19),
                               "VC" : ("Sandstorm in the vicinity",19),
                               "MI" : ("Shallow sandstorm",19),
                               "PR" : ("Partial sandstorm",19),
                               "TS" : ("Thunderous sandstorm", 19),
                               "BL" : ("Blowing sandstorm",19),
                               "DR" : ("Drifting sandstorm",19),
                               "FZ" : ("Freezing sandstorm",19),
                             }),
                      "DS" : ("Duststorm", "fog", {
                               "" : (  "Moderate duststorm",19),
                               "-" : ( "Light duststorm",19),
                               "+" : ( "Heavy duststorm",19),
                               "VC" : ("Duststorm in the vicinity",19),
                               "MI" : ("Shallow duststorm",19),
                               "PR" : ("Partial duststorm",19),
                               "TS" : ("Thunderous duststorm",19),
                               "BL" : ("Blowing duststorm",19),
                               "DR" : ("Drifting duststorm",19),
                               "FZ" : ("Freezing duststorm",19),
                             }),
                      "PO" : ("Dustwhirls", "fog", {
                               "" : (  "Moderate dustwhirls",19),
                               "-" : ( "Light dustwhirls",19),
                               "+" : ( "Heavy dustwhirls",19),
                               "VC" : ("Dustwhirls in the vicinity",19),
                               "MI" : ("Shallow dustwhirls",19),
                               "BC" : ("Patches of dustwhirls",19),
                               "PR" : ("Partial dustwhirls",19),
                               "BL" : ("Blowing dustwhirls",19),
                               "DR" : ("Drifting dustwhirls",19),
                             }),
                      "+FC" : ("Tornado", "storm", {
                               "" : (  "Moderate tornado",23),
                               "+" : ( "Raging tornado",23),
                               "VC" : ("Tornado in the vicinity",23),
                               "PR" : ("Partial tornado",23),
                               "TS" : ("Thunderous tornado",23),
                               "BL" : ("Tornado",23),
                               "DR" : ("Drifting tornado",23),
                               "FZ" : ("Freezing tornado",23),
                              }),
                      "FC" : ("Funnel cloud", "fog", {
                               "" : (  "Moderate funnel cloud",20),
                               "-" : ( "Light funnel cloud",20),
                               "+" : ( "Thick funnel cloud",20),
                               "VC" : ("Funnel cloud in the vicinity",20),
                               "MI" : ("Shallow funnel cloud",20),
                               "BC" : ("Patches of funnel cloud",20),
                               "PR" : ("Partial funnel cloud",20),
                               "BL" : ("Funnel cloud w/ wind",20),
                               "DR" : ("Drifting funnel cloud",20),
                             }),
                    }

class WeatherReport:
    """Incorporates both the unparsed textual representation of the
    weather report and the parsed values as soon as they are filled 
    in by ReportParser."""

    def _ClearAllFields(self):
        """Clear all fields values."""
        # until finished, report is invalid
        self.valid = 0
        # Clear all
        self.givenstationid=None
        self.fullreport=None
        self.temp=None
        self.windspeed=None
        self.winddir=None
        self.vis=None
        self.dewp=None
        self.humid=None
        self.press=None
        self.code=None
        self.weather=None
        self.conditions=None
        self.sky=None
        self.fulln=None
        self.cycle=None
        self.windcomp=None
        self.rtime=None
        self.pixmap=None
        self.pixmapno=None
        self.latitude=None
        self.longitude=None
        self.altitude=None
        self.stat_city=None
        self.stat_country=None
        self.reporturl=None
        self.latf=None
        self.longf=None

    def __init__(self, MetarStationCode = None):
        """Clear all fields and fill in wanted station id."""
        self._ClearAllFields()
        self.givenstationid=MetarStationCode

    def metar_to_iso8601(self, metardate) :
        """Convert a metar date to an ISO8601 date."""
        if metardate is not None:
            (date, hour, tz) = metardate.split()
            (year, month, day) = date.split('.')
            # assuming tz is always 'UTC', aka 'Z'
            return "%s-%s-%s %s:%s:00Z"%(year, month, day, hour[:2], hour[2:4])

    def getFullReport(self):
        """ Return the complete weather report.  """
        return self.fullreport

    def getTemperatureCelsius(self):
        """
        Return the temperature in degrees Celsius.
        """
        return self.temp

    def getTemperatureFahrenheit(self):
        """
        Return the temperature in degrees Fahrenheit.
        """
        return (self.temp * (9.0/5.0)) + 32.0

    def getDewPointCelsius(self):
        """
        Return dewpoint in degrees Celsius.
        """
        return self.dewp

    def getDewPointFahrenheit(self):
        """
        Return dewpoint in degrees Fahrenheit.
        """
        return (self.dewp * (9.0/5.0)) + 32.0

    def getWindSpeed(self):
        """
        Return the wind speed in meters per second.
        """
        return self.windspeed

    def getWindDirection(self):
        """
        Return wind direction in degrees.
        """
        return self.winddir

    def getWindCompass(self):
        """
        Return wind direction as compass direction
        (e.g. NE or SSE)
        """
        return self.windcomp

    def getVisibilityKilometers(self):
        """
        Return visibility in km.
        """
        return self.vis

    def getVisibilityMiles(self):
        """
        Return visibility in miles.
        """
        return self.vis / 1.609344

    def getHumidity(self):
        """
        Return relative humidity in percent.
        """
        return self.humid

    def getPressure(self):
        """
        Return pressure in hPa.
        """
        return self.press

    def getRawMetarCode(self):
        """
        Return the encoded weather report.
        """
        return self.code

    def getWeather(self):
        """
        Return short weather conditions
        """
        return self.weather

    def getConditions(self):
        return self.conditions

    def getSkyConditions(self):
        """
        Return sky conditions
        """
        return self.sky

    def getStationName(self):
        """
        Return full station name
        """
        return self.fulln

    def getStationCity(self):
        """
        Return city-part of station name
        """
        return self.stat_city

    def getStationCountry(self):
        """
        Return country-part of station name
        """
        return self.stat_country

    def getCycle(self):
        """
        Return cycle value.
        The cycle value is not the frequency or delay between 
        observations but the "time slot" in which the observation was made. 
        There are 24 cycle slots every day which usually last from N:45 to 
        N+1:45. The cycle from 23:45 to 0:45 is cycle 0.
        """
        return self.cycle

    def getStationPosition(self):
        """
        Return latitude, longitude and altitude above sea level of station
        as a tuple. Some stations don't deliver altitude, for those, None 
        is returned as altitude.  The lat/longs are expressed as follows:
        xx-yyd
        where xx is degrees, yy minutes and d the direction.
        Thus 51-14N means 51 degrees, 14 minutes north.  d may take the 
        values N, S for latitues and W, E for longitudes. Latitude and 
        Longitude may include seconds.  Altitude is always given as meters 
        above sea level, including a trailing M.
        Schipohl Int. Airport Amsterdam has, for example:
        ('52-18N', '004-46E', '-2M')
        Moenchengladbach (where I live):
        ('51-14N', '063-03E', None)
        If you need lat and long as float values, look at 
        getStationPositionFloat() instead
        """
        # convert self.altitude to string for consistency
        return (self.latitude, self.longitude, "%s"%self.altitude)

    def getStationPositionFloat(self):
        """
        Return latitude and longitude as float values in a 
        tuple (lat,long,alt).
        """
        return (self.latf,self.longf,self.altitude)

    def getStationLatitude(self) :
        """
        Return the station's latitude in dd-mm[-ss]D format :
        dd : degrees
        mm : minutes
        ss : seconds
        D : direction (N, S, E, W)
        """
        return self.latitude

    def getStationLatitudeFloat(self):
        """
        Return latitude as a float
        """
        return self.latf

    def getStationLongitude(self) :
        """
        Return the station's longitude in dd-mm[-ss]D format :
        dd : degrees
        mm : minutes
        ss : seconds
        D : direction (N, S, E, W)
        """
        return self.longitude

    def getStationLongitudeFloat(self):
        """
        Return Longitude as a float
        """
        return self.longf

    def getStationAltitude(self) :
        """
        Return the station's altitude above the sea in meters.
        """
        return self.altitude

    def getReportURL(self):
        """
        Return the URL from which the report was fetched.
        """
        return self.reporturl

    def getTime(self):
        """
        Return the time when the observation was made.  Note that this 
        is *not* the time when the report was fetched by us
        Format:  YYYY.MM.DD HHMM UTC
        Example: 2002.04.01 1020 UTC
        """
        return self.rtime

    def getISOTime(self):
        """
        Return the time when the observation was made in ISO 8601 format
        (e.g. 2002-07-25 15:12:00Z)
        """
        return(self.metar_to_iso8601(self.rtime))

    def getPixmap(self):
        """
        Return a suggested pixmap name, without extension, depending on 
        current weather.
        """
        return self.pixmap

    def getPixmapNo(self):
        return self.pixmapno


class ReportParser:
    """Parse raw METAR data from a WeatherReport object into actual 
    values and return the object with the values filled in."""

    def __init__(self, MetarReport=None):
        """Set attribute Report as specified on instantation."""
        self.Report=MetarReport

    def strreverse(self, str):
        """Reverse a string"""
        listr=list(str)
        listr.reverse()
        return "".join(listr)

    def extractCloudInformation(self) :
        """
        Extract cloud information. Return None or a tuple (sky type as a
        string of text and suggested pixmap name)
        """   
        wcloud = self.match_WeatherPart(CLOUD_RE_STR)
        if wcloud is not None :
            stype = wcloud[:3]
            if (stype == "CLR") or (stype == "SKC") :
                return ("Clear sky", "sun", 32)
            elif stype == "BKN" :
                return ("Broken clouds", "suncloud", 28)
            elif stype == "SCT" :
                return ("Scattered clouds", "suncloud", 34)
            elif stype == "FEW" :
                return ("Few clouds", "suncloud", 30)
            elif stype == "OVC" :
                return ("Overcast", "cloud", 26)
        else:
            return None # Not strictly necessary

    def extractSkyConditions(self) :
        """
        Extract sky condition information from the encoded report. Return 
        a tuple containing the description of the sky conditions as a 
        string and a suggested pixmap name for an icon representing said 
        sky condition.
        """
        wcond = self.match_WeatherPart(COND_RE_STR)
        if wcond is not None :
            if (len(wcond)>3) and (wcond.startswith('+') or wcond.startswith('-')) :
                wcond = wcond[1:]
            if wcond.startswith('+') or wcond.startswith('-') :
                pphen = 1
            elif len(wcond) < 4 :
                pphen = 0
            else :
                pphen = 2
            squal = wcond[:pphen]
            sphen = wcond[pphen : pphen + 4]
            phenomenon = _WeatherConditions.get(sphen, None)
            if phenomenon is not None :
                (name, pixmap, phenomenon) = phenomenon
                pheninfo = phenomenon.get(squal, None)
                if pheninfo is None:
                    return (name, pixmap, None)
                else:
                    return (pheninfo[0], pixmap, pheninfo[1])

    def match_WeatherPart(self, regexp) :
        """
        Return the matching part of the encoded Metar report.  
        regexp: the regexp needed to extract this part.  
        Return the first matching string or None.  
        WARNING: Some Metar reports may contain several matching 
        strings, only the first one is taken into account!
        """
        if self.Report.code is not None :
            rg = re.compile(regexp)
            for wpart in self.Report.getRawMetarCode().split() :
                match = rg.match(wpart)
                if match:
                    return match.string[match.start(0) : match.end(0)]

    def parseLatLong(self, latlong):
        """
        Parse Lat or Long in METAR notation into float values. N and E 
        are +, S and W are -. Expects one positional string and returns 
        one float value.
        """
        # I know, I could invert this if and put
        # the rest of the function into its block,
        # but I find it to be more readable this way
        if latlong is None: return None

        s=latlong.upper().strip()
        elms = s.split('-')
        ud = elms[-1][-1]
        elms[-1] = elms[-1][:-1]
        elms = map(string.atoi, elms)
        coords = 0.0
        elen = len(elms)
        if elen > 2:
            coords = coords + float(elms[2])/3600.0

        if elen > 1:
            coords = coords + float(elms[1])/60.0

        coords = coords + float(elms[0])

        if ud in ('W','S'):
            coords = -1.0*coords

        f,i = math.modf(coords)

        if elen > 2:
            f = float(fpformat.sci(f,4))
        elif elen > 1:
            f = float(fpformat.sci(f,2))
        else:
            f = 0.0

        return f+i

    def ParseReport(self, MetarReport=None):
        """Take report with raw info only and return it with in 
        parsed values filled in. Note: This function edits the
        WeatherReport object you supply!"""
        if self.Report is None and MetarReport is None:
            raise EmptyReportException, \
                "No report given on init and ParseReport()."
        elif MetarReport is not None:
            self.Report=MetarReport

        lines=self.Report.fullreport.split("\n")

        print "%s lines" % len(lines)

        for line in lines:
            try:
                header, data=line.split(":",1)
            except ValueError:
                header=data=line

            header=header.strip()
            data=data.strip()

            print header,'%%', data

            # The station id inside the report
            if header.find("("+self.Report.givenstationid+")")!=-1:
                try:
                    loc,p=data.split("(",1)
                    loc=loc.strip()
                    rloc=self.strreverse(loc)
                    rcoun,rcity=rloc.split(",",1)
                except ValueError:
                    city=""
                    coun=""
                    p=data
                try:
                    id,lat,long,ht=p.split()
                    ht=int(ht[:-1]) # cut off 'M' for meters
                except ValueError:
                    id,lat,long=p.split()
                    ht=None
                self.Report.stat_city=self.strreverse(rcity.strip())
                self.Report.stat_country=self.strreverse(rcoun.strip())
                self.Report.fulln=loc
                self.Report.latitude=lat
                self.Report.longitude=long
                self.Report.latf=self.parseLatLong(lat)
                self.Report.longf=self.parseLatLong(long)
                self.Report.altitude=ht

            # The line containing date and timr of the report

            elif (data.find("UTC"))!=-1:
                local,rt=data.split("/")
                self.Report.rtime=rt.strip()

            # temperature

            elif (header == "Temperature"):
                t,i=data.split(" ",1)
                self.Report.temp=(float(t)-32)*(5.0/9.0)

            # wind dir and speed
            
            elif (header == "Wind"):
                print header
                if (data.find("Calm")!=-1):
                    print "calm"
                    self.Report.windspeed=0.0
                    self.Report.winddir=None
                    self.Report.windcomp=None
                elif (data.find("Variable")!=-1):
                    print "var"
                    v,a,speed,r=data.split(" ",3)
                    self.Report.windspeed=(float(speed)*0.44704)
                    self.Report.winddir=None
                    self.Report.windcomp=None
                else:
                    print "elab"
                    f,t,comp,deg,r,d,speed,r=data.split(" ",7)
                    self.Report.winddir=int(deg[1:])
                    self.Report.windcomp=comp.strip()
                    self.Report.windspeed=(float(speed)*0.44704)

            # visibility

            elif (header == "Visibility"):
                for d in data.split():
                    try:
                        self.Report.vis=float(d)*1.609344
                        break
                    except ValueError:
                        pass

            # dew point
            
            elif (header == "Dew Point"):
                dp,i=data.split(" ",1)
                self.Report.dewp=(float(dp)-32)*(5.0/9.0)

            # humidity
             
            elif (header == "Relative Humidity"):
                h,i=data.split("%",1)
                self.Report.humid=int(h)

            # pressure
            
            elif (header == "Pressure (altimeter)"):
                p,r=data.split(" ",1)
                self.Report.press=(float(p)*33.863886)

            # shot weather desc. ("rain", "mist", ...)
            
            elif (header == "Weather"):
                self.Report.weather=data

            # short desc. of sky conditions

            elif (header == "Sky conditions"):
                self.Report.sky=data

            # the encoded report itself
            
            elif (header == "ob"):
                self.Report.code=data.strip()

            # the cycle value ("time slot")

            elif (header == "cycle"):
                self.Report.cycle=int(data)

        # cloud info
        cloudinfo = self.extractCloudInformation()
        if cloudinfo is not None :
            (cloudinfo, cloudpixmap, cloudpixmapno) = cloudinfo
        else :
            (cloudinfo, cloudpixmap, cloudpixmapno) = (None, None, None)
        conditions = self.extractSkyConditions()
        if conditions is not None :
            (conditions, condpixmap, condpixmapno) = conditions
        else :
            (conditions, condpixmap, condpixmapno) = (None, None, None)

        # fill the weather information
        self.Report.weather = self.Report.weather or conditions or cloudinfo
        self.Report.conditions = conditions
        # Pixmap guessed from general conditions has priority
        # over pixmap guessed from clouds
        self.Report.pixmap = condpixmap or cloudpixmap
        self.Report.pixmapno = condpixmapno or cloudpixmapno

        # report is complete
        self.Report.valid = 1

        return self.Report

class ReportFetcher:
    """Fetches a report from a given METAR id, optionally taking into
       account a different baseurl and using environment var-specified 
       proxies."""

    def __init__(self, MetarStationCode=None, baseurl="http://weather.noaa.gov/pub/data/observations/metar/decoded/"):
        """Set stationid attribute"""
        self.stationid=MetarStationCode
        self.baseurl=baseurl

    def FetchReport(self, StationCode=None):
        """Fetch a report for a given station ID from the baseurl given
        upon creation of the ReportFetcher instance."""
        if self.stationid is None and StationCode is None:
            raise EmptyIDException, \
                "No ID given on init and FetchReport()."
        elif StationCode is not None:
            self.stationid=StationCode

        self.stationid=self.stationid.upper()
        self.reporturl="%s%s.TXT" % (self.baseurl, self.stationid)
        urllib2.install_opener(
            urllib2.build_opener(urllib2.ProxyHandler, urllib2.HTTPHandler))
        try:
            fn=urllib2.urlopen(self.reporturl)
        except urllib2.HTTPError, why:
            raise NetworkException, why

        # Dump entire report in a variable
        self.fullreport=fn.read()

        if fn.info().status:
            raise NetworkException, "Could not fetch METAR report"
            
        report=WeatherReport(self.stationid)
        report.reporturl=self.reporturl
        report.fullreport=self.fullreport

        return report
        

