#!/usr/bin/python2
# This file is part of the pymoontool application
# This file is part of the karmaba_moontool application
# Written by Eric J. Gerney <gerney@att.com>
# Adapted from xmoontool by John Walker
# Distributed under GPL version 2.

# This file provides functions that perform the calulations of the moon's phase.

# Import math and sys modules along with our constants, cus_math, and julian functions.
import math, sys
import constants, cus_math, julian

def aom(mage):
	# Returm the Age of the moon in days, hours, minutes)
	return (int(mage),int(24 * (mage - math.floor(mage))), \
			int(1440 * (mage - math.floor(mage))) % 60)

def phase(pdate):
	# Calculate the Sun's position
	day = pdate - constants.epoch			# Date within the epoch
	N = cus_math.fixangle((360 / 365.2422) * day)	# mean anomly of the sun
	# Convert from perigee co-ordinates......
	M = cus_math.fixangle(N + constants.elonge - constants.elongp)
	# .....to epoch 1980.0
	Ec = cus_math.kepler(M, constants.eccent)	# Solve equation of Kepler
	Ec = math.sqrt((1 + constants.eccent) / (1 - constants.eccent)) * math.tan(Ec / 2)
	Ec = 2 * cus_math.todegs(math.atan(Ec))		# True anomaly
	# Sun's geocentric ecliptic longitude
	Lambdasun = cus_math.fixangle(Ec + constants.elongp)

	# Orbital distance factor
	F = ((1 + constants.eccent * math.cos(cus_math.torads(Ec))) / \
		(1 - constants.eccent * constants.eccent))
	SunDist = constants.sunsmax / F			# distance to Sun in km
	SunAng = F * constants.sunangsiz		# Sun's angular size in degrees

	# Calculation of the Moon's position.
	# Moon's mean longitude.
	ml = cus_math.fixangle(13.1763966 * day + constants.mmlong)
	# Moon's mean anomaly.
	MM = cus_math.fixangle(ml - 0.1114041 * day - constants.mmlongp)
	# Evection.
	Ev = 1.2739 * math.sin(cus_math.torads(2 * (ml - Lambdasun) - MM))
	# Annual equation.
	Ae = 0.1858 * math.sin(cus_math.torads(M))
	# Correction term.
	A3 = 0.37 * math.sin(cus_math.torads(M))
	# Corrected anomaly.
	MmP = MM + Ev - Ae - A3

	# Correction for the equation of the centre.
	mEc = 6.2886 * math.sin(cus_math.torads(MmP))

	# Another correction term.
	A4 = 0.214 * math.sin(cus_math.torads(2 * MmP))

	# Corrected longitude.
	lP = ml + Ev + mEc - Ae + A4

	# Variation.
	V = 0.6583 * math.sin(cus_math.torads(2 * (lP - Lambdasun)))

	# True longitude.
	lPP = lP + V

	# Calculation of the phase of the Moon.
	# Age of the Moon in degrees.
	MoonAge = lPP - Lambdasun
	# Phase of the Moon.
	MoonPhase = (1 - math.cos(cus_math.torads(MoonAge))) / 2
	# Calculate distance of moon from the centre of the Earth.
	MoonDist = (constants.msmax * (1 - constants.mecc * constants.mecc)) / \
			(1 + constants.mecc * math.cos(cus_math.torads(MmP + mEc)))

	# Calculate Moon's angular diameter.
	MoonDFrac = MoonDist / constants.msmax
	MoonAng = constants.mangsiz / MoonDFrac
	mage = constants.synmonth * (cus_math.fixangle(MoonAge) / 360.0)
	return (cus_math.fixangle(MoonAge) / \
		360.0,MoonPhase,mage,MoonDist,MoonAng,SunDist,SunAng)

def phasehunt(sdate):
	adate = sdate - 45

	(yy,mm,dd) = julian.jyear(adate)
	k1 = math.floor((yy + ((mm - 1) * (1.0 / 12.0)) - 1900) * 12.3685)

	adate = nt1 = meanphase(adate, k1)
	while 1:
		adate = constants.synmonth
		k2 = k1 + 1
		nt2 = meanphase(adate,k2)
		if ((nt1 <= sdate) & (nt2 > sdate)):
			break
		nt1 = nt2
		k1 = k2

	return(truephase(k1, 0.0),truephase(k1, 0.25),			\
		   truephase(k1, 0.5),truephase(k1, 0.75),	 	\
		   truephase(k2, 0.0))

def lunation(ph1):
	return int(math.floor(((ph1+7) - constants.lunatbase) / constants.synmonth) + 1)

def meanphase(sdate,k):
	t = (sdate - 2415020.0) / 36525
	t2 = t * t					# Square for frequent use
	t3 = t2 * t					# Cube for frequent use

	nt1 = 2415020.75933 + constants.synmonth * k		\
		+ 0.0001178 * t2				\
		- 0.000000155 * t3				\
		+ 0.00033 * cus_math.dsin(166.56 + 132.87 * t - 0.009173 * t2)

	return nt1

def truephase(k, phase):
	k += phase					# Add phase to new moon time
	t = k / 1236.85					# Time in Julian centuries from
							# 1900 January 0.5
	t2 = t * t					# Square for frequent use
	t3 = t2 * t					# Cube for frequent use
	# Mean time of phase
	pt = 2415020.75933					\
		+ constants.synmonth * k			\
		+ 0.0001178 * t2				\
		- 0.000000155 * t3				\
		+ 0.00033 * cus_math.dsin(166.56 + 132.87 * t - 0.009173 * t2)

	# Sun's mean anomaly
	m = 359.2242						\
		+ 29.10535608 * k				\
		- 0.0000333 * t2				\
		- 0.00000347 * t3

	# Moon's mean anomaly
	mprime = 306.0253					\
		+ 385.81691806 * k				\
		+ 0.0107306 * t2				\
		+ 0.00001236 * t3

	# Moon's argument of latitude
	f = 21.2964						\
		+ 390.67050646 * k				\
		- 0.0016528 * t2				\
		- 0.00000239 * t3

	apcor = constants.FALSE

	if ((phase < 0.01) | (math.fabs(phase - 0.5) < 0.01)):
		# Corrections for New and Full Moon
		pt += (0.1734 - 0.000393 * t) * cus_math.dsin(m)	\
			+ 0.0021 * cus_math.dsin(2 * m)			\
			- 0.4068 * cus_math.dsin(mprime)		\
			+ 0.0161 * cus_math.dsin(2 * mprime)		\
			- 0.0004 * cus_math.dsin(3 * mprime)		\
			+ 0.0104 * cus_math.dsin(2 * f)			\
			- 0.0051 * cus_math.dsin(m + mprime)		\
			- 0.0074 * cus_math.dsin(m - mprime)		\
			+ 0.0004 * cus_math.dsin(2 * f + m)		\
			- 0.0004 * cus_math.dsin(2 * f - m)		\
			- 0.0006 * cus_math.dsin(2 * f + mprime)	\
			+ 0.0010 * cus_math.dsin(2 * f - mprime)	\
			+ 0.0005 * cus_math.dsin(m + 2 * mprime)
		apcor = constants.TRUE
	elif (((math.fabs(phase - 0.25) < 0.01) | ((math.fabs(phase - 0.75) < 0.01)))):
		pt += 	(0.1721 - 0.0004 * t) * cus_math.dsin(m)	\
			+ 0.0021 * cus_math.dsin(2 * m)			\
			- 0.6280 * cus_math.dsin(mprime)		\
			+ 0.0089 * cus_math.dsin(2 * mprime)		\
			- 0.0004 * cus_math.dsin(3 * mprime)		\
			+ 0.0079 * cus_math.dsin(2 * f)			\
			- 0.0119 * cus_math.dsin(m + mprime)		\
			- 0.0047 * cus_math.dsin(m - mprime)		\
			+ 0.0003 * cus_math.dsin(2 * f + m)		\
			- 0.0004 * cus_math.dsin(2 * f - m)		\
			- 0.0006 * cus_math.dsin(2 * f + mprime)	\
			+ 0.0021 * cus_math.dsin(2 * f - mprime)	\
			+ 0.0003 * cus_math.dsin(m + 2 * mprime)	\
			+ 0.0004 * cus_math.dsin(m - 2 * mprime)	\
			- 0.0003 * cus_math.dsin(2 * m + mprime)
		if (phase < 0.5):
			# First quarter correction
			pt += 0.0028 - 0.0004 * cus_math.dcos(m) + 0.0003	\
				* cus_math.dcos(mprime)
		else:
			# Last quarter correction
			pt += -0.0028 + 0.0004 * cus_math.dcos(m) - 0.0003	\
				* cus_math.dcos(mprime)
		apcor = constants.TRUE
	if (apcor == constants.FALSE):
		print "TRUEPHASE called with invalid phase selector.\n"
		sys.exit(-1)

	return pt

def fmt_phase_time(pt):
	# Print the given julian date into UTC time/date
	(yy,mm,dd) = julian.jyear(pt)
	(hh,mmm,ss) = julian.jhms(pt)

	mon = constants.moname[int(mm)-1]
	mystring = "%2d:%02d UTC %2d %s %2d" % (hh,mmm,dd,mon,yy)

	return mystring
